# mka78
# Adds A78 headers to Atari 7800 ROMs.
# Copyright (C) 2004 by T. Mathys (killer@vantage.ch).
#
#
# mka78 is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# A78 header format:
#
#       Offset  Size (Bytes)    Description
#       ----------------------------------------------------------------------
#       0       1               Header version (1 ?)
#       1       16              "ATARI7800", zero terminated/padded
#       17      32              Cart title, zero terminated/padded
#       49      4               Cart size in bytes (big endian)
#       53      2               Cart type (big endian)
#       55      1               Controller 1 type
#       56      1               Controller 2 type
#       57      1               TV type: 0 = NTSC, 1 = PAL
#       58      42              Unused (all bytes should be zero)
#       100     28              "ACTUAL CART DATA STARTS HERE", not terminated
#
#       Cart type
#       ---------
#       Bit 0           Pokey cart
#       Bit 1           Supercart bank switched
#       Bit 2           Supercart RAM at $4000
#       Bit 3           ROM at $4000
#       Bit 4           Bank 6 at $4000
#       Bit 8-15        Special (0 = Normal cart)
#
#       Controller type
#       ---------------
#       0               None
#       1               Joystick
#       2               Light gun
#
require "readbytes"
require "getoptlong"

MKA78_VERSION = "0.99"

HEADER_VERSION = 1
DEVICE_NAME = "ATARI7800\0\0\0\0\0\0\0"
MAX_CART_TITLE_LENGTH = 31

UNUSED_SIZE = 42
CART_MARKER = "ACTUAL CART DATA STARTS HERE"

CARTTYPE_POKEY				= 0x0001
CARTTYPE_SUPERCART		= 0x0002
CARTTYPE_SUPERCARTRAM	= 0x0004
CARTTYPE_ROM4000			= 0x0008
CARTTYPE_BANK64000		= 0x0010

TVTYPE_NTSC = 0
TVTYPE_PAL = 1

CTRLTYPE_NONE = 0
CTRLTYPE_JOYSTICK = 1
CTRLTYPE_LIGHTGUN = 2


def printWarning(message)
	$stderr.puts "Warning : #{message}"
end


class CmdLineParser

	attr_reader :cartType
	attr_reader :inputFileName
	attr_reader :leftController
	attr_reader :outputFileName
	attr_reader :pal
	attr_reader :rightController
	attr_reader :title

	def initialize
		@cartType = 0
		@inputFileName = nil
		@leftController = CTRLTYPE_NONE
		@outputFileName = nil
		@pal = false
		@rightController = CTRLTYPE_NONE
		@title = ""
	end
	
	def CmdLineParser.displayHelp
		puts <<blah
Usage: mka78.rb file [options]

--bank6, -6
    Cart has bank 6 mapped in at 0x4000.
--left-controller=<controller>, -1<controller>
--right-controller=<controller>, -2<controller>
    Set controller type. <controller> can be "none", "joystick"
    or "lightgun" (without quotes). Abbreviations can be used,
    so for instance "j" or "lightg" will work aswell.
    Default controller type is "none".
--output-file=<filename>, -o<filename>
    Set output filename. If no output filename is specified,
    the output file's name will be the name of the input file,
    with the extension .a78 appended.
--pal, -p
    PAL cart.
--pokey, -y
    Pokey cart.
--rom, -m
    Cart has ROM mapped in at 0x4000.
--supercart, -s
    Cart uses supercart bankswitching.
--supercart-ram, -r
    Cart has supercart RAM mapped in at 0x4000.
--title, -t
    Cart title. Can be at most #{MAX_CART_TITLE_LENGTH} characters, longer titles
    will be truncated.
blah
	end
	
	# Parse the command line.
	# Raises a StandardError when an error is encountered.
	# In this case, the values returned by the attribute
	# readers should be considered invalid.
	def parse
	
		opts = GetoptLong.new(
			["--bank6", "-6", GetoptLong::NO_ARGUMENT],
			["--left-controller", "-1", GetoptLong::REQUIRED_ARGUMENT],
			["--output-file", "-o", GetoptLong::REQUIRED_ARGUMENT],
			["--pal", "-p", GetoptLong::NO_ARGUMENT],
			["--pokey", "-y", GetoptLong::NO_ARGUMENT],
			["--right-controller", "-2", GetoptLong::REQUIRED_ARGUMENT],
			["--rom", "-m", GetoptLong::NO_ARGUMENT],
			["--supercart", "-s", GetoptLong::NO_ARGUMENT],
			["--supercart-ram", "-r", GetoptLong::NO_ARGUMENT],
			["--title", "-t", GetoptLong::REQUIRED_ARGUMENT]
		)
		opts.quiet = true
		
		opts.each do |opt, arg|
			case opt
				when "--bank6"
					@cartType |= CARTTYPE_BANK64000
				when "--left-controller"
					@leftController = getController(arg)
				when "--output-file"
					@outputFileName = arg
				when "--pal"
					@pal = true
				when "--pokey"
					@cartType |= CARTTYPE_POKEY
				when "--right-controller"
					@rightController = getController(arg)
				when "--rom"
					@cartType |= CARTTYPE_ROM4000
				when "--supercart"
					@cartType |= CARTTYPE_SUPERCART
				when "--supercart-ram"
					@cartType |= CARTTYPE_SUPERCARTRAM
				when "--title"
					@title = arg
			end
		end
		
		# Exactly one input file given ?
		if 0 == ARGV.length then
			raise StandardError, "No input file given"
		end
		if 1 < ARGV.length then
			raise StandardError, "More than one input file given"
		end
		@inputFileName = ARGV[0]
		
		# Save/create output file name
		if nil == @outputFileName then
			@outputFileName = @inputFileName + ".a78"
		end
	end
	
	
private

	# Takes a string and tries to convert it to a CTRLTYPE_xxx constant.
	def getController(s)
	
		tmp = s.downcase
	
		return CTRLTYPE_NONE if 0 == "none".index(tmp)
		
		return CTRLTYPE_JOYSTICK if 0 == "joystick".index(tmp)
		
		return CTRLTYPE_LIGHTGUN if 0 == "lightgun".index(tmp)

		raise StandardError, "Unknown controller type : #{s}"
	end	
end


# Writes an A78 file
#
# cartdata	:	Cart data (String)
# clp		:	CmdLineParser with parameters
#
def writeA78(cartdata, clp)
	File.open(clp.outputFileName, "wb") do |file|

		file.putc(HEADER_VERSION)
		file.print(DEVICE_NAME)
		
		# Clamp or pad cart title if necessary and write to file
		t = clp.title
		if t.length > MAX_CART_TITLE_LENGTH then
			printWarning("Cart title is longer than #{MAX_CART_TITLE_LENGTH} characters and will be truncated.")
			t = t[0, MAX_CART_TITLE_LENGTH]
		else
			t = t + "\000"*(MAX_CART_TITLE_LENGTH-t.length)
		end
		file.print(t)
		file.putc(0)	# Write terminating zero
		
		# Cart size
		size = cartdata.length
		file.putc((size >> 24) & 255)
		file.putc((size >> 16) & 255)
		file.putc((size >> 8) & 255)
		file.putc(size & 255)

		# Cart type
		file.putc((clp.cartType >> 8) & 255)
		file.putc(clp.cartType & 255)

		# Controller types
		file.putc(clp.leftController & 255)
		file.putc(clp.rightController & 255)
		
		# TV type
		file.putc(clp.pal ? TVTYPE_PAL : TVTYPE_NTSC)

		# Pad header and write cartridge marker
		UNUSED_SIZE.times {file.putc(0)}
		file.print(CART_MARKER)

		file.print(cartdata)
	end
end

	
begin
	puts "mka78 #{MKA78_VERSION} - Copyright (C) 2004 by Thomas Mathys."
	puts "This program is distributed under the terms of the GPL."
	puts "See mka78.rb for more information."

	if 0 == ARGV.length then
		puts
		CmdLineParser.displayHelp
		exit(0)
	end
	
	clp = CmdLineParser.new
	clp.parse

	# Load input file	
	cartdata = nil
	File.open(clp.inputFileName, "rb") do |file|
		cartdata = file.readbytes(File.size(clp.inputFileName))
	end
	
	writeA78(cartdata, clp)

	puts "#{clp.outputFileName} written"
	
rescue StandardError => e
	$stderr.puts "Error : #{e}"
	exit(1)
end
